@@ -1,13 +1,11 @@
-import uuid
 import logging
 import re
-import datetime
 import os
+from subprocess import Popen, PIPE
 import sys
 
-from twisted.internet import defer
-
 from tron import node, command_context
+from tron.filehandler import FileHandleManager
 from tron.utils import timeutils
 from tron.utils import state
 
@@ -97,20 +95,17 @@ class ActionRunContext(object):
             raise KeyError(name)
 
 
-# States where we have executed the run.
-# ACTION_RUN_EXECUTED_STATES = [ACTION_RUN_FAILED, ACTION_RUN_SUCCEEDED]
-
-
 class ActionRun(object):
     """An instance of running a action"""
     STATE_CANCELLED = state.NamedEventState('cancelled')
-    STATE_UNKNOWN = state.NamedEventState('unknown')
+    STATE_UNKNOWN = state.NamedEventState('unknown', short_name='UNKWN')
     STATE_FAILED = state.NamedEventState('failed')
     STATE_SUCCEEDED = state.NamedEventState('succeeded')
     STATE_RUNNING = state.NamedEventState('running')
-    STATE_STARTING = state.NamedEventState('starting')
+    STATE_STARTING = state.NamedEventState('starting', short_chars=5)
     STATE_QUEUED = state.NamedEventState('queued')
     STATE_SCHEDULED = state.NamedEventState('scheduled')
+    STATE_SKIPPED = state.NamedEventState('skipped')
 
     STATE_SCHEDULED['ready'] = STATE_QUEUED
     STATE_SCHEDULED['queue'] = STATE_QUEUED
@@ -134,11 +129,21 @@ class ActionRun(object):
         event_state['success'] = STATE_SUCCEEDED
         event_state['fail'] = STATE_FAILED
 
+    # We can tell tron to skip a failed step
+    STATE_FAILED['skip'] = STATE_SKIPPED
+    STATE_CANCELLED ['skip'] = STATE_SKIPPED
+
     STATE_QUEUED['schedule'] = STATE_SCHEDULED
 
-    def __init__(self, action, context=None, output_path=None):
+    def __init__(self,
+        action,
+        context=None,
+        output_path=None,
+        node=None,
+        id=None,
+    ):
         self.action = action
-        self.id = None
+        self.id = id
 
         self.run_time = None    # What time are we supposed to start
         self.start_time = None  # What time did we start
@@ -146,7 +151,7 @@ class ActionRun(object):
         self.exit_status = None
         self.machine = state.StateMachine(ActionRun.STATE_SCHEDULED)
 
-        self.node = None
+        self.node = node
 
         if context is None:
             # Provide dummy values for context variables that JobRun provides
@@ -193,17 +198,18 @@ class ActionRun(object):
         return self.tail_file(self.stderr_path, num_lines)
 
     def tail_file(self, path, num_lines):
-        try:
-            out = open(path, 'r')
-            out.close()
-        except IOError:
+        if not path or not os.path.exists(path):
             return []
-
         if not num_lines or num_lines <= 0:
             num_lines = sys.maxint
 
-        tail = os.popen("tail -n %s %s" % (num_lines + 1, path))
-        lines = tail.read().split('\n')[:-1]
+        try:
+            cmd = ('tail', '-n', str(num_lines), path)
+            tail_sub = Popen(cmd, stdout=PIPE)
+            lines = [line.rstrip() for line in tail_sub.stdout]
+        except OSError:
+            log.error("Could not tail %s." % path)
+            return []
 
         if len(lines) > num_lines:
             lines[0] = "..."
@@ -213,11 +219,15 @@ class ActionRun(object):
     def attempt_start(self):
         self.machine.transition('ready')
 
-        if all([r.is_success for r in self.required_runs]):
-            self.start()
+        if all(r.is_success or r.is_skipped for r in self.required_runs):
+            return self.start()
+
+    def check_state(self, state):
+        """Check if the state machine can be transitioned to state."""
+        return self.machine.check(state)
 
     def start(self):
-        if self.state not in (self.STATE_QUEUED, self.STATE_SCHEDULED):
+        if not self.machine.check('start'):
             raise InvalidStartStateError(self.state)
 
         log.info("Starting action run %s", self.id)
@@ -225,46 +235,52 @@ class ActionRun(object):
         self.end_time = None
         self.machine.transition('start')
         assert self.state == self.STATE_STARTING, self.state
-        self._open_output_file()
 
         if not self.is_valid_command:
             log.error("Command for action run %s is invalid: %r",
                       self.id, self.action.command)
-            self.fail(-1)
-            return
+            return self.fail(-1)
 
         # And now we try to actually start some work....
-        self.action_command = ActionCommand(self.id,
-                                            self.command,
-                                            stdout=self.stdout_file,
-                                            stderr=self.stderr_file)
+        self._setup_output_files()
+        self.action_command = ActionCommand(
+            self.id,
+            self.command,
+            stdout=self.stdout_file,
+            stderr=self.stderr_file
+        )
         self.action_command.machine.listen(True, self._handle_action_command)
         try:
             df = self.node.run(self.action_command)
             df.addErrback(self._handle_errback)
         except node.Error, e:
             log.warning("Failed to start %s: %r", self.id, e)
+            return False
+        return True
+
+    def _setup_output_files(self):
+        file_manager = FileHandleManager.get_instance()
+        self.stdout_file = file_manager.open(self.stdout_path)
+        self.stderr_file = file_manager.open(self.stderr_path)
 
     def cancel(self):
-        self.machine.transition('cancel')
+        return self.machine.transition('cancel')
 
     def schedule(self):
-        self.machine.transition('schedule')
+        return self.machine.transition('schedule')
 
     def queue(self):
-        self.machine.transition('queue')
+        return self.machine.transition('queue')
 
-    def _open_output_file(self):
-        try:
-            log.info("Opening file %s for output", self.stdout_path)
-            if self.stdout_path:
-                self.stdout_file = open(self.stdout_path, 'a')
-            if self.stderr_path:
-                self.stderr_file = open(self.stderr_path, 'a')
-        except IOError, e:
-            log.error(str(e) + " - Not storing command output!")
+    def skip(self):
+        """Mark the run as having been skipped."""
+        if self.machine.transition('skip'):
+            self.start_dependants()
+            return True
+        return False
 
     def _close_output_file(self):
+        """Attempt to close any open file handlers."""
         if self.stdout_file:
             self.stdout_file.close()
         if self.stderr_file:
@@ -282,6 +298,7 @@ class ActionRun(object):
         generated.
         """
         log.info("Action error: %s", str(result))
+        self._close_output_file()
         if isinstance(result.value, node.ConnectError):
             log.warning("Failed to connect to host %s for run %s",
                         self.node.hostname, self.id)
@@ -301,22 +318,25 @@ class ActionRun(object):
         """
         log.debug("Action command state change: %s", self.action_command.state)
         if self.action_command.state == ActionCommand.RUNNING:
-            self.machine.transition('started')
-        elif self.action_command.state == ActionCommand.FAILSTART:
+            return self.machine.transition('started')
+
+        if self.action_command.state == ActionCommand.FAILSTART:
             self._close_output_file()
-            self.fail(None)
-        elif self.action_command.state == ActionCommand.EXITING:
+            return self.fail(None)
+
+        if self.action_command.state == ActionCommand.EXITING:
             if self.action_command.exit_status is None:
-                self.fail_unknown()
-            elif self.action_command.exit_status == 0:
-                self.succeed()
-            else:
-                self.fail(self.action_command.exit_status)
-        elif self.action_command.state == ActionCommand.COMPLETE:
+                return self.fail_unknown()
+            if not self.action_command.exit_status:
+                return self.succeed()
+            return self.fail(self.action_command.exit_status)
+
+        if self.action_command.state == ActionCommand.COMPLETE:
             self._close_output_file()
-        else:
-            raise Error("Invalid state for action command : %r" %
-                        self.action_command)
+            return
+
+        raise Error(
+            "Invalid state for action command : %r" % self.action_command)
 
     def start_dependants(self):
         for run in self.waiting_runs:
@@ -327,13 +347,14 @@ class ActionRun(object):
             log.info("Not running waiting run %s, the dependant action failed",
                      run.id)
 
-    def fail(self, exit_status):
+    def fail(self, exit_status=0):
         """Mark the run as having failed, providing an exit status"""
         log.info("Action run %s failed with exit status %r",
                  self.id, exit_status)
-        self.machine.transition('fail')
-        self.exit_status = exit_status
-        self.end_time = timeutils.current_time()
+        if self.machine.transition('fail'):
+            self.exit_status = exit_status
+            self.end_time = timeutils.current_time()
+            return True
 
     def fail_unknown(self):
         """Mark the run as having failed, but note that we don't actually know
@@ -341,21 +362,24 @@ class ActionRun(object):
         """
         log.info("Lost communication with action run %s", self.id)
 
-        self.machine.transition('fail_unknown')
-        self.exit_status = None
-        self.end_time = None
+        if self.machine.transition('fail_unknown'):
+            self.exit_status = None
+            self.end_time = None
+            return True
 
     def mark_success(self):
         self.exit_status = 0
         self.end_time = timeutils.current_time()
-        self.machine.transition('success')
+        return self.machine.transition('success')
 
     def succeed(self):
         """Mark the run as having succeeded"""
         log.info("Action run %s succeeded", self.id)
 
-        self.mark_success()
-        self.start_dependants()
+        if self.mark_success():
+            self.start_dependants()
+            return True
+        return False
 
     def restore_state(self, state_data):
         self.id = state_data['id']
@@ -369,18 +393,44 @@ class ActionRun(object):
         # We were running when the state file was built, so we have no idea
         # what happened now.
         if self.is_running:
-            self.machine.transition('fail_unknown')
+            return self.machine.transition('fail_unknown')
 
     @property
     def data(self):
-        return {'id': self.id,
-                'state': str(self.state),
-                'run_time': self.run_time,
-                'start_time': self.start_time,
-                'end_time': self.end_time,
-                'command': self.command
+        """This data is used to serialize the state of this action run."""
+        return {
+            'id':           self.id,
+            'state':        str(self.state),
+            'run_time':     self.run_time,
+            'start_time':   self.start_time,
+            'end_time':     self.end_time,
+            'command':      self.command,
         }
 
+    def repr_data(self, max_lines=None):
+        """Return a dictionary that represents the external view of this
+        action run.
+        """
+        data = {
+            'id':           self.id,
+            'name':         self.action.name,
+            'state':        self.state.short_name,
+            'node':         self.node.hostname,
+            'command':      self.command,
+            'raw_command':  self.action.command,
+            'run_time':     self.run_time,
+            'start_time':   self.start_time,
+            'end_time':     self.end_time,
+            'exit_status':  self.exit_status,
+            'requirements': [req.name for req in self.action.required_actions],
+            'stdout':       [],
+            'stderr':       []
+        }
+        if max_lines:
+            data['stdout'] = self.tail_stdout(max_lines)
+            data['stderr'] = self.tail_stderr(max_lines)
+        return data
+
     def render_command(self):
         """Render our configured command under the command context.
 
@@ -411,6 +461,11 @@ class ActionRun(object):
         except Exception:
             return False
 
+    @property
+    def is_done(self):
+        return self.state in (self.STATE_FAILED, self.STATE_SUCCEEDED,
+                              self.STATE_CANCELLED, self.STATE_SKIPPED)
+
     @property
     def is_queued(self):
         return self.state == self.STATE_QUEUED
@@ -423,11 +478,6 @@ class ActionRun(object):
     def is_scheduled(self):
         return self.state == self.STATE_SCHEDULED
 
-    @property
-    def is_done(self):
-        return self.state in (self.STATE_FAILED, self.STATE_SUCCEEDED,
-                              self.STATE_CANCELLED)
-
     @property
     def is_unknown(self):
         return self.state == self.STATE_UNKNOWN
@@ -448,38 +498,57 @@ class ActionRun(object):
     def is_success(self):
         return self.state == self.STATE_SUCCEEDED
 
+    @property
+    def is_skipped(self):
+        return self.state == self.STATE_SKIPPED
+
 
 class Action(object):
-    def __init__(self, name=None):
+    def __init__(self, name=None, command=None, node_pool=None,
+                 required_actions=None):
         self.name = name
+        self.command = command
+        self.node_pool = node_pool
 
-        self.required_actions = []
+        self.required_actions = required_actions or []
         self.job = None
-        self.command = None
+
+    @classmethod
+    def from_config(cls, config, node_pools):
+        return cls(
+            name=config.name,
+            command=config.command,
+            node_pool=node_pools[config.node] if config.node else None
+        )
 
     def __eq__(self, other):
-        if not isinstance(other, Action) \
-           or self.name != other.name \
-           or self.command != other.command:
+        if (not isinstance(other, Action)
+           or self.name != other.name
+           or self.command != other.command):
             return False
 
-        return all([me == you for (me, you) in zip(self.required_actions,
-                                                   other.required_actions)])
+        return all(me == you for (me, you) in zip(self.required_actions,
+                                                   other.required_actions))
 
     def __ne__(self, other):
         return not self == other
 
-    def build_run(self, job_run):
-        """Build an instance of ActionRun for this action
-
-        This is used by the scheduler when scheduling a run
+    def build_run(self, job_run, cleanup=False):
+        """Build an instance of ActionRun for this action. If cleanup=True
+        we're building a cleanup action run.
         """
-        new_run = ActionRun(self, context=job_run.context)
+        callback = job_run.cleanup_action_run if cleanup else job_run.run_completed
 
-        new_run.id = "%s.%s" % (job_run.id, self.name)
-        new_run.output_path = job_run.output_path
-
-        return new_run
+        action_run = ActionRun(
+            self,
+            context=job_run.context,
+            node=job_run.node,
+            id="%s.%s" % (job_run.id, self.name),
+            output_path=job_run.output_path)
+        action_run.machine.listen(True, job_run.job.notify)
+        action_run.machine.listen(ActionRun.STATE_SUCCEEDED, callback)
+        action_run.machine.listen(ActionRun.STATE_FAILED,    callback)
+        return action_run
 
 
 class ActionCommand(object):
@@ -528,12 +597,12 @@ class ActionCommand(object):
 
     def started(self):
         self.start_time = timeutils.current_timestamp()
-        self.machine.transition("start")
+        return self.machine.transition("start")
 
     def exited(self, exit_status):
         self.end_time = timeutils.current_timestamp()
         self.exit_status = exit_status
-        self.machine.transition("exit")
+        return self.machine.transition("exit")
 
     def write_stderr(self, value):
         if self.stderr_file:
@@ -544,7 +613,7 @@ class ActionCommand(object):
             self.stdout_file.write(value)
 
     def write_done(self):
-        self.machine.transition("close")
+        return self.machine.transition("close")
 
     def __repr__(self):
         return "[ActionCommand %s] %s : %s" % (self.id, self.command,
