@@ -4,7 +4,10 @@ import shutil
 from collections import deque
 
 from tron import action, command_context, event
+from tron.action import Action, ActionRun
+from tron.scheduler import scheduler_from_config, ConstantScheduler
 from tron.utils import timeutils
+from tron.utils.proxy import CollectionProxy
 
 
 class Error(Exception):
@@ -44,6 +47,30 @@ class JobRun(object):
             self, parent=self.job.event_recorder)
         self.event_recorder.emit_info("created")
 
+        # Setup proxies
+        self.proxy_action_runs_with_cleanup = CollectionProxy(
+            self.action_runs_with_cleanup, [
+                ('is_failure',      any,    False),
+                ('is_starting',     any,    False),
+                ('is_running',      any,    False),
+                ('is_scheduled',    any,    False),
+                ('is_unknown',      any,    False),
+                ('is_queued',       all,    False),
+                ('is_cancelled',    all,    False),
+                ('is_skipped',      all,    False),
+                ('is_done',         all,    False),
+                ('check_state',     all,    True),
+                ('cancel',          all,    True),
+                ('succeed',         all,    True),
+                ('fail',            any,    True)
+        ])
+
+        self.proxy_action_runs = CollectionProxy(
+            self.action_runs, [
+                ('schedule',        all,    True),
+                ('queue',           all,    True),
+        ])
+
     @property
     def output_path(self):
         return os.path.join(self.job.output_path, self.id)
@@ -51,8 +78,8 @@ class JobRun(object):
     def set_run_time(self, run_time):
         self.run_time = run_time
 
-        for action in self.action_runs_with_cleanup:
-            action.run_time = run_time
+        for action_run in self.action_runs_with_cleanup:
+            action_run.run_time = run_time
 
     def seconds_until_run_time(self):
         run_time = self.run_time
@@ -81,8 +108,28 @@ class JobRun(object):
                             " - cancelling", self.id)
                 self.cancel()
 
+    @property
+    def state(self):
+        """The overall state of this job run. Based on the state of its actions.
+        """
+        if self.is_success:
+            return ActionRun.STATE_SUCCEEDED
+        if self.is_cancelled:
+            return ActionRun.STATE_CANCELLED
+        if self.is_running:
+            return ActionRun.STATE_RUNNING
+        if self.is_failure:
+            return ActionRun.STATE_FAILED
+        if self.is_scheduled:
+            return ActionRun.STATE_SCHEDULED
+        if self.is_queued:
+            return ActionRun.STATE_QUEUED
+        if self.is_skipped:
+            return ActionRun.STATE_SKIPPED
+        return ActionRun.STATE_UNKNOWN
+
     def start(self):
-        if not (self.is_scheduled or self.is_queued):
+        if not self.proxy_action_runs_with_cleanup.perform('check_state', 'start'):
             raise InvalidStartStateError("Not scheduled")
 
         log.info("Starting action job %s", self.id)
@@ -92,6 +139,11 @@ class JobRun(object):
         try:
             for action_run in self.action_runs:
                 action_run.attempt_start()
+            if self.cleanup_action_run:
+                # Don't call attempt_start or it will just go. The JobRun
+                # object will kick this when it's actually time to run (any
+                # failure or total success).
+                self.cleanup_action_run.machine.transition('ready')
             self.event_recorder.emit_info("started")
         except action.Error, e:
             log.warning("Failed to start actions: %r", e)
@@ -118,14 +170,9 @@ class JobRun(object):
             except Error, e:
                 log.warning("Attempt to start failed: %r", e)
 
-    def last_success_check(self):
-        if (not self.job.last_success or self.run_num >
-                self.job.last_success.run_num):
-            self.job.last_success = self
-
     def run_completed(self):
         if self.is_success:
-            self.last_success_check()
+            self.job.update_last_success(self)
 
             if self.job.constant and self.job.enabled:
                 self.job.build_run().start()
@@ -152,17 +199,32 @@ class JobRun(object):
 
     @property
     def data(self):
-        data = {'runs': [a.data for a in self.action_runs],
-                'cleanup_run': None,
-                'run_num': self.run_num,
-                'run_time': self.run_time,
-                'start_time': self.start_time,
-                'end_time': self.end_time
+        """This data is used to serialize the state of this job run."""
+        data = {
+            'id':           self.id,
+            'runs':         [a.data for a in self.action_runs],
+            'cleanup_run':  None,
+            'run_num':      self.run_num,
+            'run_time':     self.run_time,
+            'start_time':   self.start_time,
+            'end_time':     self.end_time
         }
         if self.cleanup_action_run is not None:
             data['cleanup_run'] = self.cleanup_action_run.data
         return data
 
+    def repr_data(self):
+        """A dict that represents the externally visible state."""
+        return {
+            'id':           self.id,
+            'state':        self.state.short_name,
+            'node':         self.node.hostname if self.node else None,
+            'run_num':      self.run_num,
+            'run_time':     self.run_time,
+            'start_time':   self.start_time,
+            'end_time':     self.end_time,
+        }
+
     def restore(self, data):
         self.start_time = data['start_time']
         self.end_time = data['end_time']
@@ -176,30 +238,27 @@ class JobRun(object):
 
         self.event_recorder.emit_info("restored")
 
-    def schedule(self):
-        for r in self.action_runs:
-            r.schedule()
+    def __getattr__(self, name):
+        # The order here is important.  We don't want to raise too many
+        # exceptions, so proxies should be ordered by those most likely
+        # to be used.
+        for proxy in [
+            self.proxy_action_runs_with_cleanup,
+            self.proxy_action_runs
+        ]:
+            try:
+                return proxy.perform(name)
+            except AttributeError:
+                pass
 
-    def queue(self):
-        for r in self.action_runs:
-            r.queue()
-
-    def cancel(self):
-        for r in self.action_runs_with_cleanup:
-            r.cancel()
-
-    def succeed(self):
-        for r in self.action_runs_with_cleanup:
-            r.mark_success()
-
-    def fail(self):
-        for r in self.action_runs_with_cleanup:
-            r.fail(0)
+        # We want to re-raise this exception because the proxy code
+        # will not be relevant in the stack trace
+        raise AttributeError(name)
 
     @property
     def action_runs_with_cleanup(self):
         if self.cleanup_action_run is not None:
-            return [a for a in self.action_runs] + [self.cleanup_action_run]
+            return self.action_runs + [self.cleanup_action_run]
         else:
             return self.action_runs
 
@@ -207,58 +266,29 @@ class JobRun(object):
     def should_start(self):
         if not self.job.enabled or self.is_running:
             return False
-        return (self.job.next_to_finish(self.node
-                                        if self.job.all_nodes
-                                        else None)
-                is self)
-
-    @property
-    def is_failure(self):
-        return any([r.is_failure for r in self.action_runs_with_cleanup])
+        node = self.node if self.job.all_nodes else None
+        return self.job.next_to_finish(node) is self
 
     @property
     def all_but_cleanup_success(self):
-        return all([r.is_success for r in self.action_runs])
+        """Overloaded all_but_cleanup_success, because we can still succeed
+        if some actions were skipped.
+        """
+        return all(r.is_success or r.is_skipped for r in self.action_runs)
 
     @property
     def is_success(self):
-        return all([r.is_success for r in self.action_runs_with_cleanup])
+        """Overloaded is_success, because we can still succeed if some
+        actions were skipped.
+        """
+        return all(
+            r.is_success or r.is_skipped for r in self.action_runs_with_cleanup)
 
     @property
     def all_but_cleanup_done(self):
         if self.is_failure:
             return True
-        return not any(r.is_running or r.is_queued or r.is_scheduled
-                       for r in self.action_runs)
-
-    @property
-    def is_done(self):
-        return not any([r.is_running or r.is_queued or r.is_scheduled
-                        for r in self.action_runs_with_cleanup])
-
-    @property
-    def is_queued(self):
-        return all([r.is_queued for r in self.action_runs_with_cleanup])
-
-    @property
-    def is_starting(self):
-        return any([r.is_starting for r in self.action_runs_with_cleanup])
-
-    @property
-    def is_running(self):
-        return any([r.is_running for r in self.action_runs_with_cleanup])
-
-    @property
-    def is_scheduled(self):
-        return any([r.is_scheduled for r in self.action_runs_with_cleanup])
-
-    @property
-    def is_unknown(self):
-        return any([r.is_unknown for r in self.action_runs_with_cleanup])
-
-    @property
-    def is_cancelled(self):
-        return all([r.is_cancelled for r in self.action_runs_with_cleanup])
+        return all(r.is_done for r in self.action_runs)
 
     @property
     def cleanup_job_status(self):
@@ -280,43 +310,93 @@ class Job(object):
 
     run_num = 0
 
+    # Constants for Job status
+    STATUS_DISABLED =   "DISABLED"
+    STATUS_ENABLED =    "ENABLED"
+    STATUS_UNKNOWN =    "UNKNOWN"
+    STATUS_RUNNING =    "RUNNING"
+
     def next_num(self):
         self.run_num += 1
         return self.run_num - 1
 
-    def __init__(self, name=None, action=None, context=None,
-                 event_recorder=None):
+    def __init__(self,
+        name=None,
+        action=None,
+        context=None,
+        event_recorder=None,
+        queueing=True,
+        run_limit=RUN_LIMIT,
+        all_nodes=False,
+        scheduler=None,
+        node_pool=None,
+        enabled=True
+    ):
         self.name = name
         self.topo_actions = [action] if action else []
         self.cleanup_action = None
-        self.scheduler = None
+        self.scheduler = scheduler
         self.runs = deque()
 
-        self.queueing = True
-        self.all_nodes = False
-        self.enabled = True
-        self.constant = False
+        self.queueing = queueing
+        self.all_nodes = all_nodes
+        self.enabled = enabled
         self.last_success = None
 
-        self.run_limit = RUN_LIMIT
-        self.node_pool = None
+        self.run_limit = run_limit
+        self.node_pool = node_pool
         self.output_path = None
         self.state_callback = lambda: None
         self.context = command_context.CommandContext(self, context)
         self.event_recorder = event.EventRecorder(self, parent=event_recorder)
 
+    @classmethod
+    def from_config(cls, job_config, node_pools, time_zone):
+        """Build a job from a ConfigJob."""
+        job = cls(
+            name=job_config.name,
+            queueing=job_config.queueing,
+            run_limit=job_config.run_limit,
+            all_nodes=job_config.all_nodes,
+            node_pool=node_pools[job_config.node] if job_config.node else None,
+            scheduler=scheduler_from_config(job_config.schedule, time_zone),
+            enabled=job_config.enabled
+        )
+
+        new_actions = dict(
+            (name, Action.from_config(action_conf, node_pools))
+            for name, action_conf in job_config.actions.iteritems()
+        )
+
+        for action in new_actions.values():
+            for dep in job_config.actions[action.name].requires:
+                action.required_actions.append(new_actions[dep])
+            job.add_action(action)
+
+        if job_config.cleanup_action:
+            action = Action.from_config(job_config.cleanup_action, node_pools)
+            job.register_cleanup_action(action)
+
+        return job
+
     def _register_action(self, action):
         """Prepare an action to be *owned* by this job"""
         if action in self.topo_actions:
-            raise Error("Action %s already in jobs %s" %
-                        (action.name, job.name))
+            raise Error("Action %s already in jobs %s" % (
+                action.name, self.name))
+
+    def register_cleanup_action(self, action):
+            self.cleanup_action = action
+            action.job = self
+            self._register_action(action)
 
     def listen(self, spec, callback):
         """Mimic the state machine interface for listening to events"""
         assert spec is True
         self.state_callback = callback
 
-    def _notify(self):
+    def notify(self):
+        """Used as a callback for state machine state changes."""
         self.state_callback()
 
     def add_action(self, action):
@@ -378,13 +458,7 @@ class Job(object):
 
     def newest_run_by_state(self, state):
         for run in self.runs:
-            if state == 'SUCC' and run.is_success or \
-                state == 'CANC' and run.is_cancelled or \
-                state == 'RUNN' and run.is_running or \
-                state == 'FAIL' and run.is_failure or \
-                state == 'SCHE' and run.is_scheduled or \
-                state == 'QUE' and run.is_queued or \
-                state == 'UNKWN' and run.is_unknown:
+            if run.state.short_name == state:
                 return run
 
         log.warning("No runs with state %s exist", state)
@@ -434,30 +508,27 @@ class Job(object):
 
         return self.scheduler.next_runs(self)
 
-    def _make_action_run(self, job_run, action_inst, callback):
-            action_run = action_inst.build_run(job_run)
-
-            action_run.node = job_run.node
-
-            action_run.machine.listen(True, self._notify)
-            action_run.machine.listen(action.ActionRun.STATE_SUCCEEDED,
-                                      callback)
-            action_run.machine.listen(action.ActionRun.STATE_FAILED,
-                                      callback)
-
-            return action_run
+    def update_last_success(self, run):
+        """Update the last_success run if the run number is greater then the
+        previous last_success."""
+        assert run.job == self
+        if not self.last_success or run.run_num > self.last_success.run_num:
+            self.last_success = run
 
     def build_action_dag(self, job_run, all_actions):
         """Build actions and setup requirements"""
-        action_runs_by_name = {}
-        for action_inst in all_actions:
-            action_run = self._make_action_run(job_run,
-                                               action_inst,
-                                               job_run.run_completed)
 
-            action_runs_by_name[action_inst.name] = action_run
+        def create_action_inst(action_inst):
+            action_run = action_inst.build_run(job_run)
             job_run.action_runs.append(action_run)
+            return action_run
 
+        action_runs_by_name = dict(
+            (action_inst.name, create_action_inst(action_inst))
+            for action_inst in all_actions
+        )
+
+        for action_inst in all_actions:
             for req_action in action_inst.required_actions:
                 if req_action.name not in action_runs_by_name:
                     raise ConfigBuildMismatchError(
@@ -465,11 +536,13 @@ class Job(object):
                         req_action.name)
 
                 # Two-way, waiting runs and required_runs
-                action = action_runs_by_name[req_action.name]
-                action.waiting_runs.append(action_run)
-                action_run.required_runs.append(action)
+                req_action_run = action_runs_by_name[req_action.name]
+                waiting_action_run = action_runs_by_name[action_inst.name]
+                req_action_run.waiting_runs.append(waiting_action_run)
+                waiting_action_run.required_runs.append(req_action_run)
 
-    def build_run(self, node=None, actions=None, run_num=None):
+    def build_run(self, node=None, actions=None, run_num=None,
+                  cleanup_action=None):
         job_run = JobRun(self, run_num=run_num)
 
         job_run.node = node or self.node_pool.next()
@@ -488,9 +561,10 @@ class Job(object):
 
         self.build_action_dag(job_run, actions)
 
-        if self.cleanup_action is not None:
-            cleanup_action_run = self._make_action_run(
-                job_run, self.cleanup_action, job_run.cleanup_completed)
+        cleanup_action = cleanup_action or self.cleanup_action
+        if cleanup_action is not None:
+            cleanup_action_run = self.cleanup_action.build_run(
+                job_run, cleanup=True)
             job_run.cleanup_action_run = cleanup_action_run
 
         return job_run
@@ -531,17 +605,43 @@ class Job(object):
 
     @property
     def data(self):
-        return {'runs': [r.data for r in self.runs],
-                'enabled': self.enabled
+        """This data is used to serialize the state of this job."""
+        return {
+            'runs': [r.data for r in self.runs],
+            'enabled': self.enabled
         }
 
+    def repr_data(self):
+        """Returns a dict that is the external representation of this job."""
+        last_success = self.last_success.end_time if self.last_success else None
+        return {
+            'name':         self.name,
+            'scheduler':    str(self.scheduler),
+            'action_names': [a.name for a in self.topo_actions],
+            'node_pool':    [n.hostname for n in self.node_pool.nodes],
+            'status':       self.status,
+            'last_success': last_success,
+        }
+
+    @property
+    def status(self):
+        """The Jobs current status is determined by its last/next run."""
+        current_run = self.next_to_finish()
+        if not current_run:
+            return self.STATUS_DISABLED
+        if current_run.is_running:
+            return self.STATUS_RUNNING
+        if current_run.is_scheduled:
+            return self.STATUS_ENABLED
+        return self.STATUS_UNKNOWN
+
     def restore(self, data):
         self.enabled = data['enabled']
 
         for r_data in data['runs']:
             try:
                 self.restore_run(r_data)
-            except job.Error, e:
+            except Error, e:
                 log.warning("Failed to restore job: %r (%r)", r_data, e)
                 continue
 
@@ -557,7 +657,11 @@ class Job(object):
 
         action_list = filter(action_filter, self.topo_actions)
 
-        run = self.build_run(run_num=data['run_num'], actions=action_list)
+        ca = (self.cleanup_action
+              if self.cleanup_action and action_filter(self.cleanup_action)
+              else None)
+        run = self.build_run(run_num=data['run_num'], actions=action_list,
+                             cleanup_action=ca)
         self.run_num = max([run.run_num + 1, self.run_num])
 
         run.restore(data)
@@ -568,5 +672,15 @@ class Job(object):
 
         return run
 
+    def setup_job_dir(self, working_dir):
+        """Setup a directory to store this jobs logs."""
+        self.output_path = os.path.join(working_dir, self.name)
+        if not os.path.exists(self.output_path):
+            os.mkdir(self.output_path)
+
+    @property
+    def constant(self):
+        return isinstance(self.scheduler, ConstantScheduler)
+
     def __str__(self):
         return "JOB:%s" % self.name
