@@ -1,17 +1,25 @@
 from __future__ import with_statement
 import logging
+import logging.handlers
 import os
 import shutil
-import subprocess
-import sys
+import socket
 import time
-import weakref
 import yaml
 
-import tron
-from tron import job, config, command_context, event
+from twisted.conch.client.options import ConchOptions
 from twisted.internet import reactor
-from tron.utils import timeutils
+
+import tron
+from tron import command_context
+from tron import event
+from tron import monitor
+from tron.config import config_parse
+from tron.config.config_parse import ConfigError
+from tron.job import Job
+from tron.node import Node, NodePool
+from tron.service import Service
+from tron.utils import timeutils, emailer
 
 
 log = logging.getLogger('tron.mcp')
@@ -33,6 +41,11 @@ class UnsupportedVersionError(Error):
     pass
 
 
+class ConfigApplyError(Exception):
+    """Errors during config application"""
+    pass
+
+
 class StateHandler(object):
 
     def __init__(self, mcp, working_dir, writing=False):
@@ -199,21 +212,38 @@ class MasterControlProgram(object):
     def __init__(self, working_dir, config_file, context=None):
         self.jobs = {}
         self.services = {}
-        self.nodes = []
+
+        # Mapping of node name to node (or node pool) objects
+        self.nodes = {}
+
+        # Path to the config file
         self.config_file = config_file
-        self.context = context
+
+        # Root command context
+        self.context = context or command_context.CommandContext()
+
         self.monitor = None
+
+        # Time zone of the system clock
         self.time_zone = None
+
+        # Record events for the entire system. Child event recorders may record
+        # events for specific jobs, job runs, actions, action runs, etc. and
+        # these events will be propagated up but not down the event recorder
+        # tree.
         self.event_recorder = event.EventRecorder(self)
+
+        # Control writing of the state file
         self.state_handler = StateHandler(self, working_dir)
 
         root = logging.getLogger('')
         self.base_logging_handlers = list(root.handlers)
 
+    ### CONFIGURATION ###
+
     def live_reconfig(self):
         self.event_recorder.emit_info("reconfig")
         try:
-
             # Temporarily disable state writing because reconfig can cause a
             # lot of state changes
             old_state_writing = self.state_handler.writing_enabled
@@ -223,7 +253,7 @@ class MasterControlProgram(object):
 
             # Any new jobs will need to be scheduled
             self.run_jobs()
-        except Exception, e:
+        except Exception:
             self.event_recorder.emit_critical("reconfig_failure")
             log.exception("Reconfig failure")
             raise
@@ -232,10 +262,8 @@ class MasterControlProgram(object):
 
     def load_config(self):
         log.info("Loading configuration from %s" % self.config_file)
-        opened_config = open(self.config_file, "r")
-        configuration = config.load_config(opened_config)
-        configuration.apply(self)
-        opened_config.close()
+        with open(self.config_file, 'r') as f:
+            self.apply_config(config_parse.load_config(f))
 
     def config_lines(self):
         try:
@@ -255,15 +283,175 @@ class MasterControlProgram(object):
         except IOError, e:
             log.error(str(e) + " - Cannot write to configuration file!")
 
-    def setup_job_dir(self, job):
-        job.output_path = os.path.join(self.state_handler.working_dir,
-                                       job.name)
-        if not os.path.exists(job.output_path):
-            os.mkdir(job.output_path)
+    def apply_config(self, conf, skip_env_dependent=False):
+        """Apply a configuration. If skip_env_dependent is True we're
+        loading this locally to test the config as part of tronfig. We want to
+        skip applying some settings because the local machine we're using to
+        edit the config may not have the same environment as the live
+        trond machine.
+        """
+        self._apply_working_directory(conf.working_dir)
+        self._apply_loggers(conf.syslog_address)
 
+        if not skip_env_dependent:
+            ssh_options = self._ssh_options_from_config(conf.ssh_options)
+        else:
+            ssh_options = config_parse.valid_ssh_options({})
+
+        self.context.base = conf.command_context
+        self._apply_nodes(conf.nodes, ssh_options)
+        self._apply_node_pools(conf.node_pools)
+
+        self.time_zone = conf.time_zone
+        self._apply_jobs(conf.jobs)
+        self._apply_services(conf.services)
+        self._apply_notification_options(conf.notification_options)
+
+    def _apply_working_directory(self, wd):
+        """Apply working directory.  If wd=None try os.environ['TMPDIR'].
+        If that doesn't exist, use /tmp.
+        """
+        if self.state_handler.working_dir:
+            if self.state_handler.working_dir != wd:
+                log.warn('trond must be restarted for changes to the working'
+                         ' directory to take effect.')
+            return
+
+        if not wd:
+            if 'TMPDIR' in os.environ:
+                wd = os.environ['TMPDIR']
+            else:
+                wd = '/tmp'
+
+        if not os.path.isdir(wd):
+            raise ConfigApplyError("Specified working directory '%s' is not"
+                                   " a directory" % wd)
+
+        if not os.access(wd, os.W_OK):
+            raise ConfigApplyError("Specified working directory '%s' is not"
+                                   " writable" % wd)
+
+        self.state_handler.working_dir = wd
+
+    def _apply_loggers(self, syslog_address):
+        root = logging.getLogger('')
+        handlers_to_be_removed = set(h for h in root.handlers
+                                     if h not in self.base_logging_handlers)
+
+        # Only change handlers if they will actually be different from the old
+        # handlers
+        new_handlers = []
+        if syslog_address:
+            if not isinstance(syslog_address, basestring):
+                self.syslog_address = tuple(self.syslog_address)
+
+            try:
+                h = logging.handlers.SysLogHandler(syslog_address)
+                fmt_str = "tron[%(process)d]: %(message)s"
+                h.setFormatter(logging.Formatter(fmt_str))
+                new_handlers.append(h)
+            except socket.error:
+                raise ConfigError('%s is not a valid syslog address' %
+                                  syslog_address)
+
+        for h in handlers_to_be_removed:
+            log.info('Removing logging handler %s', h)
+            root.removeHandler(h)
+
+        for h in new_handlers:
+            log.info('Adding logging handler %s', h)
+            root.addHandler(h)
+
+    def _ssh_options_from_config(self, ssh_conf):
+        ssh_options = ConchOptions()
+        if ssh_conf.agent:
+            if 'SSH_AUTH_SOCK' in os.environ:
+                ssh_options['agent'] = True
+            else:
+                raise ConfigError("No SSH Agent available ($SSH_AUTH_SOCK)")
+        else:
+            ssh_options['noagent'] = True
+
+        for file_name in ssh_conf.identities:
+            file_path = os.path.expanduser(file_name)
+            if not os.path.exists(file_path):
+                raise ConfigError("Private key file '%s' doesn't exist" %
+                                  file_name)
+            if not os.path.exists(file_path + ".pub"):
+                raise ConfigError("Public key '%s' doesn't exist" %
+                                  (file_name + ".pub"))
+
+            ssh_options.opt_identity(file_name)
+
+        return ssh_options
+
+    def _apply_nodes(self, node_confs, ssh_options):
+        self.nodes = dict(
+            (name, Node.from_config(config, ssh_options))
+            for name, config in node_confs.iteritems()
+        )
+
+    def _apply_node_pools(self, pool_confs):
+        self.nodes.update(
+            (name, NodePool.from_config(config, self.nodes))
+            for name, config in pool_confs.iteritems()
+        )
+
+    def _apply_jobs(self, job_configs):
+        """Add and remove jobs based on the configuration."""
+        for job_config in job_configs.values():
+            log.debug("Building new job %s", job_config.name)
+            job = Job.from_config(job_config, self.nodes, self.time_zone)
+            self.add_job(job)
+
+        for job_name in (set(self.jobs.keys()) - set(job_configs.keys())):
+            log.debug("Removing job %s", job_name)
+            self.remove_job(job_name)
+
+    def _apply_services(self, srv_configs):
+        """Add and remove services."""
+
+        services_to_add = []
+        for srv_config in srv_configs.values():
+            log.debug("Building new services %s", srv_config.name)
+            service = Service.from_config(srv_config, self.nodes)
+            services_to_add.append(service)
+
+        for srv_name in (set(self.services.keys()) - set(srv_configs.keys())):
+            log.debug("Removing service %s", srv_name)
+            self.remove_service(srv_name)
+
+        # Go through our constructed services and add them. We'll catch all the
+        # failures and throw an exception at the end if anything failed. This
+        # is a mitigation against a bug easily cause us to be in an
+        # inconsistent state, probably due to bad code elsewhere.
+        # TODO: what actually causes this
+        failure = False
+        for service in services_to_add:
+            try:
+                self.add_service(service)
+            except Exception, e:
+                log.exception("Failed adding new service.", e)
+                failure = e
+
+        if failure:
+            raise ConfigError("Failed adding services %s" % failure)
+
+    def _apply_notification_options(self, notification_conf):
+        if notification_conf is not None:
+            if self.monitor:
+                self.monitor.stop()
+
+            em = emailer.Emailer(notification_conf.smtp_host,
+                                 notification_conf.notification_addr)
+            self.monitor = monitor.CrashReporter(em, self)
+            self.monitor.start()
+
+    ### JOBS ###
     def add_job(self, job):
         if job.name in self.services:
             raise ValueError("Job %s is already a service", job.name)
+
         if job.name in self.jobs:
 
             # Jobs have a complex eq implementation that allows us to catch
@@ -286,13 +474,17 @@ class MasterControlProgram(object):
 
         self.jobs[job.name] = job
 
-        # update time zone information in scheduler to match config
-        if job.scheduler is not None:
-            job.scheduler.time_zone = self.time_zone
-
+        # add some command context variables
         job.set_context(self.context)
+
         job.event_recorder.set_parent(self.event_recorder)
-        self.setup_job_dir(job)
+
+        # make the directory for output
+        job.setup_job_dir(self.state_handler.working_dir)
+
+        # Tell the job to call store_state() whenever its state changes.
+        # job isn't actaully a StateMachine object, but its interface tries
+        # to look like one.
         job.listen(True, self.state_handler.store_state)
 
     def remove_job(self, job_name):
@@ -300,9 +492,10 @@ class MasterControlProgram(object):
             raise ValueError("Job %s unknown", job_name)
 
         job = self.jobs.pop(job_name)
-
         job.disable()
 
+    ### SERVICES ###
+
     def add_service(self, service):
         if service.name in self.jobs:
             raise ValueError("Service %s is already a job", service.name)
@@ -332,6 +525,8 @@ class MasterControlProgram(object):
         service = self.services.pop(service_name)
         service.stop()
 
+    ### OTHER ACTIONS ###
+
     def _schedule(self, run):
         secs = run.seconds_until_run_time()
         if secs == 0:
@@ -340,6 +535,7 @@ class MasterControlProgram(object):
 
     def schedule_next_run(self, job):
         if job.runs and job.runs[0].is_scheduled:
+            log.info('Job is already scheduled')
             return
 
         for next in job.next_runs():
