@@ -16,33 +16,16 @@ class CircularTransitionError(Error):
 log = logging.getLogger(__name__)
 
 
-class FunctionItemState(object):
-
-    # Rules should be a list of tuples in the form (next_state,
-    # validation_funtion) where if the function returns true, the next state is
-    # valid.
-
-    rules = list()
-
-    def __getitem__(self, key):
-        next_states = [state for state, func in self.rules if func(key)]
-        if len(next_state) > 1:
-            raise InvalidRuleError("Too many results")
-
-        if next_states:
-            return next_states[0]
-        else:
-            raise KeyError()
-
-
 class NamedEventState(dict):
     """Simple state type that allows you to easily use a dictionary for a state
     implementation. A raw dictionary works fine as well, but this might be more
     clear, plus it gives you a name.
     """
 
-    def __init__(self, name, **kwargs):
+    def __init__(self, name, short_name=None, short_chars=4, **kwargs):
         self.name = name
+        self._short_name = short_name
+        self._short_chars = short_chars
         super(NamedEventState, self).__init__(**kwargs)
 
     def __eq__(self, other):
@@ -51,12 +34,24 @@ class NamedEventState(dict):
         except AttributeError:
             return False
 
+    def __nonzero__(self):
+        return bool(self.name)
+
     def __str__(self):
         return self.name
 
     def __repr__(self):
         return "<%r %s>" % (self.__class__.__name__, self.name)
 
+    @property
+    def short_name(self):
+        """If a short_name was given use that, otherwise return the first
+        self._short_letters letters of the name in caps.
+        """
+        if self._short_name:
+            return self._short_name
+        return self.name[:self._short_chars].upper()
+
 
 def named_event_by_name(starting_state, state_name):
     """Traverse the state graph and pull out the one with the provided name.
@@ -94,6 +89,14 @@ class StateMachine(object):
         self._listeners = list()
         self._state_by_name = None
 
+
+    def check(self, target):
+        """Check if the state can be transitioned to target. Returns the
+        destination state if target is a valid state to transition to,
+        None otherwise.
+        """
+        return self.state.get(target, None)
+
     def transition(self, target, stop_item=None):
         """Check our current state for a transition based on the input 'target'
 
@@ -103,9 +106,8 @@ class StateMachine(object):
         """
         log.debug("Checking for transition from %r (%r)", self.state, target)
 
-        try:
-            next_state = self.state[target]
-        except KeyError:
+        next_state = self.check(target)
+        if next_state is None:
             return False
 
         prev_state = self.state
@@ -131,7 +133,7 @@ class StateMachine(object):
 
     def listen(self, listen_spec, callback):
 
-        """Listen for the specific state of set of states and callback on the
+        """Listen for the specific state or set of states and callback on the
         function provided.
 
         The callback is called AFTER the state transition has been made.
