@@ -1,21 +1,47 @@
-from types import FunctionType
+from contextlib import contextmanager
 import functools
 import itertools
 import logging
+from types import FunctionType
 
-from testify import *
+from testify import assert_not_reached, turtle, TestCase, class_teardown
 from testify.test_case import TwistedFailureError
 from twisted.internet import reactor, defer
 from twisted.python import failure
 
-from tron.utils import twistedutils
-
-
 log = logging.getLogger(__name__)
 
 _waiting = False
 
 
+# this exists as logging.NullHandler as of Python 2.7
+class NullHandler(logging.Handler):
+    def emit(self, record):
+        pass
+
+
+@contextmanager
+def no_handlers_for_logger(name=None):
+    """Temporarily remove handlers all handlers from a logger.
+
+    Use this in a `with` block. For example::
+
+        with no_handlers_for_logger('tron.mcp'):
+            # do stuff with mcp
+
+    Any handlers you add inside the `with` block will be removed at the end.
+    """
+    log = logging.getLogger(name)
+    old_handlers = log.handlers
+
+    # add null handler so logging doesn't yell about there being no handlers
+    log.handlers = [NullHandler()]
+
+    yield
+
+    log.handlers = old_handlers
+
+
 class ReactorTestCase(TestCase):
     """Inherit from this subclass if any of your suite's tests use the Twisted
     reactor, so that the reactor isn't stopped until your tests run.
@@ -98,7 +124,7 @@ def wait_for_deferred(deferred, timeout=None):
     deferred.addErrback(on_failure)
 
     if timeout is not None:
-        reactor.callLater(timeout, stop)
+        reactor.callLater(timeout, reactor.stop)
 
     try:
         deferred.addBoth(append)
@@ -119,14 +145,14 @@ def wait_for_deferred(deferred, timeout=None):
         _waiting = False
         results = None
         if failures:
-            failure = failures[0]
+            first_failure = failures[0]
 
             # By this point we've already lost too much of our exception
             # information (traceback, stack) to re-create the real exception.
             # So what we'll have to do is hope our test framework can handle
             # twisted failure objects so we can get some useful information out
             # of them.
-            raise TwistedFailureError(failure)
+            raise TwistedFailureError(first_failure)
             #raise failure.type, failure.value, failure.getTracebackObject()
             #failures[0].raiseException()
 
@@ -168,7 +194,7 @@ def run_reactor(timeout=DEFAULT_TIMEOUT, assert_raises=None):
 
             found_error = False
             try:
-                result = wait_for_deferred(deferred)
+                wait_for_deferred(deferred)
             except TwistedFailureError, e:
                 if assert_raises:
                     d_fail = e.args[0]
